""" Test the FDTD detectors """

## Imports
import fdtd
from fdtd.backend import backend as bd
from fdtd.waveforms import normalized_gaussian_pulse
from fdtd.conversions import *
from fdtd.sources import SoftArbitraryPointSource
from fdtd.detectors import CurrentDetector, BlockDetector
from fdtd.constants import X, Y, Z

import pytest
from fixtures import grid, pml, periodic_boundary
import matplotlib.pyplot as plt

## Tests

x_ = 4
y_ = 4
z_ = 4

# theory-driven development? there must be a more efficient way to test this kind of code...

# if/when detectors get polarizations,
# test in three axes - spin the grid to make sure everything's consistent

# each detector starts out with all the parameters empty ([])
# this causes edge cases when no time steps have passed, and should get explicit test cases


# @pytest.mark.skipif(skip_test_backend, reason="PyTorch is not installed.")
def test_DC_absorbing_current_detector_all_bends(grid, pml, backends):
    fdtd.set_backend(backends)

    # we maybe want our test cases to be orthogonal:
    # The test case with the patch antenna also relies on other modules like
    # the electric field detector & PEC, seems like a smaller test could be in order

    # However, because of the overlapping object issue, seems like this'll be harder.

    # This doesn't seem to work properly, not sure why-
    # see issue #11, will look into it.

    conductivity = 1e9
    # absorb = bd.ones((grid.Nx,grid.Ny,grid.Nz))
    # absorb[x_,y_,z_] = 0
    grid[x_ - 1 : x_ + 1, y_ - 1 : y_ + 1, z_ - 1 : z_ + 1] = fdtd.AbsorbingObject(
        permittivity=1.0, conductivity=conductivity
    )

    # AbsorbingObject doesn't create a magnetic field?
    # How could that ever cause a current density?

    # grid[x_,y_,z_] = fdtd.PointSource(period=500)
    cd = fdtd.CurrentDetector(name="Dave")
    grid[x_, y_, z_] = cd
    grid[x_:x_, y_:y_, z_:z_] = fdtd.BlockDetector()
    # FIXME: AbsorbingObject must be added last for some reason,

    n = 1000
    # grid.run(total_time=n)
    for _ in range(n):
        grid.update_E()
        grid.E[x_, y_, z_, Z] = normalized_gaussian_pulse(
            grid.time_passed,
            (n / 10.0) * grid.time_step,
            center=(n / 2) * grid.time_step,
        )
        grid.update_H()
        grid.time_steps_passed += 1

    #
    current_time_history = bd.array(grid.detectors[0].I).reshape(n)
    electric_field_time_history = bd.array(grid.detectors[1].E).reshape(n, 3)[:, Z]
    # print(grid.E[x_,y_,z_,Z])
    #
    # print(current_time_history)
    voltage = electric_field_time_history * grid.grid_spacing
    resistivity = 1.0 / conductivity
    resistance = (resistivity * grid.grid_spacing) / (
        grid.grid_spacing * grid.grid_spacing
    )
    expected_current = voltage / resistance
    # print(expected_current)
    # print(current_time_history)
    # plt.plot(current_time_history)
    # plt.show()
    # plt.plot(electric_field_time_history)


def test_CurrentDetector_shape(grid):
    edetector = BlockDetector()
    cdetector = CurrentDetector()
    grid[4, 4, 4] = cdetector
    grid[4, 4, 5] = edetector
    grid.run(10)
    # at the moment, edetector has an extra dimension for polarization
    # which isn't included in cdetector.
    assert bd.array(cdetector.I).shape[0:2] == bd.array(edetector.E).shape[0:2]
    assert bd.array(cdetector.I).shape[0] == 10


def test_SAPS_detector_register(grid):
    detector = SoftArbitraryPointSource(bd.zeros(1), impedance=50.0)
    grid[4, 4, 4] = detector
    grid.run(100)
    assert isinstance(grid.detectors[0], CurrentDetector)



""" Test the FDTD boundaries """

## Imports

# 3rd party
import pytest

# fdtd
import fdtd

# fixtures
from fixtures import grid, pml, periodic_boundary
from fdtd.backend import backend as bd

from fdtd.boundaries import DomainBorderPML
## Tests


def test_periodic_boundary_in_grid_boundary_list(grid, periodic_boundary):
    grid[0, :, :] = periodic_boundary
    assert periodic_boundary in grid.boundaries


def test_periodic_boundary_raises_error_when_indexed_with_slice(
    grid, periodic_boundary
):
    with pytest.raises(IndexError):
        grid[0:2, :, :] = periodic_boundary


def test_periodic_boundary_placed_in_middle_of_grid(grid, periodic_boundary):
    with pytest.raises(IndexError):
        grid[2, :, :] = periodic_boundary


def test_pml_in_grid_boundary_list(grid, pml):
    grid[0:3, :, :] = pml
    assert pml in grid.boundaries


def test_pml_placed_in_middle_of_grid(grid, pml):
    with pytest.raises(IndexError):
        grid[2:4, :, :] = pml

def test_pml_in_grid_boundary_list(grid, pml):
    grid[0:3, :, :] = pml
    assert pml in grid.boundaries

def test_DomainPML_exception(grid, pml):
    with pytest.raises(IndexError):
        DomainBorderPML(grid, grid.Nx//2+1)

def test_DomainPML_success(grid, pml):
    DomainBorderPML(grid, 3)

def test_DomainPML_bonehead_indexing():
    #make sure I did the indexing properly!
    #no overlaps.
    test_grid = bd.zeros((9,9,9))
    border_cells = 4

    #top and bottom
    test_grid[:, : ,0:border_cells] += 1
    test_grid[:, : ,-border_cells:] += 1

    #left and right
    test_grid[0:border_cells, :, border_cells:-border_cells] += 1
    test_grid[-border_cells:, :, border_cells:-border_cells] += 1

    #front and back
    test_grid[border_cells:-border_cells, 0:border_cells, border_cells:-border_cells] += 1
    test_grid[border_cells:-border_cells, -border_cells:, border_cells:-border_cells] += 1

    comparison_grid = bd.ones((9,9,9))
    comparison_grid[4,4,4] = 0

    assert (test_grid == comparison_grid).all()

# test non-unique names



""" Test the FFT system """

## Imports
import fdtd
import pytest
from fixtures import grid, pml, periodic_boundary

from fdtd.fourier import FrequencyRoutines
from fdtd.backend import backend as bd
from fdtd.detectors import LineDetector, BlockDetector, CurrentDetector
from fdtd.sources import SoftArbitraryPointSource
import numpy as np

## Tests


def test_padding(grid):
    fr = FrequencyRoutines(grid, None)
    input_data = bd.zeros((1000))
    dt = 1e-9
    #sample rate 1 GHz
    #1 microsecond capture length
    #total length
    # freq_window = ()


    required_padding, end_time = \
            fr.compute_padding(input_data, dt=dt, freq_window_tuple=None, fft_num_bins_in_window=None,
                                        fft_bin_freq_resolution=None, )

    # assert end_time == pytest.approx(2.3, rel=0.1)

                        #test that end_time is time_passed
    # assert end_time == pytest.approx(2.3, rel=0.1)


def test_FFT_single_detector_empty_time(grid):
    detector = BlockDetector()
    grid[4,4,4] = detector
    fr = FrequencyRoutines(grid, detector)
    fr.FFT()

def test_FFT_single_detector(grid):
    detector = BlockDetector()
    grid[4,4,4] = detector
    grid.run(100)
    # detector.E[:][0][0][0] = np.sin().tolist()
    # tolist required for accurate test of existing multiple-dimension
    fr = FrequencyRoutines(grid, detector)
    spectrum_freqs, spectrum = fr.FFT()


def test_SAPS_impedance(grid):
    detector = SoftArbitraryPointSource(np.zeros(1),impedance = 50.0)
    grid[4,4,4] = detector
    grid.run(100)
    fr = FrequencyRoutines(grid, detector)
    fr.impedance()



# use QUCS to generate something 3 port with known S-params



""" Test the FDTD objects """


## Imports
import fdtd
import pytest

## Tests


#
# def test_object_grid_overlap(grid, pml):
#     grid[x_-1:x_+1,y_-1:y_+1,z_-1:z_+1] = fdtd.AbsorbingObject(permittivity=1.0,
#     conductivity=conductivity)
#     # grid[x_,y_,z_] = fdtd.PointSource(period=500)
#



""" Test the FFT system """

## Imports
import fdtd
import pytest
from fixtures import grid, pml, periodic_boundary
from fdtd.boundaries import DomainBorderPML
from fdtd.objects import Object, AbsorbingObject
from fdtd.waveforms import normalized_gaussian_pulse
from fdtd.sources import SoftArbitraryPointSource
from fdtd.fourier import FrequencyRoutines
from fdtd.backend import backend as bd
import numpy as np
import matplotlib.pyplot as plt


def create_patch_antenna(normalized_probe_position):
    # [Samaras 2004]
    # Probe fed microstrip antenna
    # probe in middle and translated along X
    # designed f = 2.3 GHz,
    # excited with gaussian pulse
    # centered at 2.3 GHz, harmonics up to 4 GHz
    # 6-layer thick PML
    # A non-uniform grid was used in the Samaras' model.
    # The maximum cell size was 2 mm.
    # other tests should be physically smaller if possible;
    # performance limitation here seems to be
    # the discretization size, not the timestep
    border_cells = 6
    patch_width =  59.4e-3 # m
    patch_length = 40.4e-3 # m
    substrate_thickness = 1.27e-3 # m
    dielectric_constant = 2.42 # m
    loss_tangent = 0.0019 #



    N_substrate_cells = 3
    spacing = substrate_thickness / N_substrate_cells

    # 1 / (((1.27e-3 / 10)  meters) / c) ≈ 100x f, A-ok

    Nx = int(patch_width/spacing)+(2*border_cells)
    Ny = int(patch_length/spacing)+(2*border_cells)
    Nz = int(substrate_thickness/spacing)+(2*border_cells) + 2 # finite thickness of copper planes

    grid = fdtd.Grid(
        shape=(Nx, Ny, Nz),
        grid_spacing=spacing,
        permittivity=1.0,
        permeability=1.0,
        courant_number=None,  # calculate the courant number
    )


    pulse_fwhm = 2.3e-9

    simulation_steps = int(5 * (pulse_fwhm / grid.time_step))
    pulse_center_time = (simulation_steps * grid.time_step) / 5.0

    DomainBorderPML(grid, border_cells=border_cells)

    sl = slice(border_cells, -border_cells)

    copper_object = AbsorbingObject(permittivity=1.0, conductivity=1e8)


    grid[sl, sl, border_cells:border_cells+1] = copper_object # ground plane

    bottom_copper_plane = border_cells
    top_copper_plane = (border_cells+1+N_substrate_cells)

    grid[sl, sl, top_copper_plane:top_copper_plane+1] = copper_object #top plane

    probe_Nx = int(patch_width/spacing)//2 + border_cells
    probe_Ny = int(normalized_probe_position*(int(patch_length/spacing)//2)) + border_cells

    times = np.arange(0,simulation_steps) * grid.time_step

    waveform_array = normalized_gaussian_pulse(times, pulse_fwhm, center=pulse_center_time)

    source = SoftArbitraryPointSource(waveform_array=waveform_array, impedance=50.0)

    grid[probe_Nx,probe_Ny,border_cells+1] = source

    grid[probe_Nx:probe_Nx+1, probe_Ny:probe_Ny+1,
                        border_cells+2:top_copper_plane+1] = copper_object # probe feed via

    substrate = Object(permittivity=dielectric_constant)
    grid[border_cells:probe_Nx, sl, bottom_copper_plane+1:top_copper_plane] = substrate
    grid[probe_Nx+1:-border_cells, sl, bottom_copper_plane+1:top_copper_plane] = substrate

    grid[sl, border_cells:probe_Ny, bottom_copper_plane+1:top_copper_plane] = substrate
    grid[sl, probe_Ny+1:-border_cells, bottom_copper_plane+1:top_copper_plane] = substrate
    # don't overlap with the source port!

    return grid, source, simulation_steps
    #


# use pytest -s here to see live output
@pytest.mark.slow
def test_antenna_impedance():
    fdtd.set_backend("torch.cuda.float32")
    # a very slow test
    distance_from_edge = bd.array([0.0,0.25,0.35,0.50,0.67,1.00])
    samaras_FDTD_R = bd.array([169,136,106,72,32,0])

    grid, source, simulation_steps = create_patch_antenna(0)
    for idx, d in enumerate(distance_from_edge):
        grid, source, _ = create_patch_antenna(d)

        grid.run(simulation_steps)

        fr = FrequencyRoutines(grid, source)
        fr.impedance()

# use QUCS to generate something 3 port with known S-params

#ask floris if there're any test cases for optical



""" Test the FDTD sources """


## Imports
import fdtd
import pytest

from fixtures import grid, pml, periodic_boundary
from fdtd.sources import SoftArbitraryPointSource
from fdtd.sources import PlaneSource
## Tests


def test_SoftArbitraryPointSource(grid):
    pass

def test_PlaneSource_polarization_error(grid):
    with pytest.raises(ValueError):
        grid[0, :, :] = PlaneSource(polarization='x')
    with pytest.raises(ValueError):
        grid[:, 0, :] = PlaneSource(polarization='y')
    with pytest.raises(ValueError):
        grid[:, :, 0] = PlaneSource(polarization='z')

def test_PlaneSource_polarization_inference(grid):
    ps = PlaneSource(polarization='y')
    grid[0, :, :] = ps
    assert ps._Epol == 1 and ps._Hpol == 2
    ps = PlaneSource(polarization='z')
    grid[0, :, :] = ps
    assert ps._Epol == 2 and ps._Hpol == 1
    ps = PlaneSource(polarization='x')
    grid[:, 0, :] = ps
    assert ps._Epol == 0 and ps._Hpol == 2
    ps = PlaneSource(polarization='z')
    grid[:, 0, :] = ps
    assert ps._Epol == 2 and ps._Hpol == 0
    ps = PlaneSource(polarization='x')
    grid[:, :, 0] = ps
    assert ps._Epol == 0 and ps._Hpol == 1
    ps = PlaneSource(polarization='y')
    grid[:, :, 0] = ps
    assert ps._Epol == 1 and ps._Hpol == 0




import pytest

import fdtd
from fdtd.backend import backend as bd
import numpy as np
from fdtd.waveforms import normalized_gaussian_pulse
# Thought for sure these wouldn't need testing, but nooo
def test_waveforms_all_bends(backends):
    fdtd.set_backend(backends)
    times = np.arange(0,1000) / 1000.0
    waveform_array = normalized_gaussian_pulse(times, 0.01,center =0.5)
    assert np.max(waveform_array) == pytest.approx(1.0, rel=0.1)
    assert waveform_array[0] == pytest.approx(0.0, rel=1e-6)
    assert waveform_array[-1] == pytest.approx(0.0, rel=1e-6)



""" Test the FDTD backends """

## Imports

# 3rd party
import pytest

# This library
import fdtd

try:
    from fdtd.backend import NumpyBackend, TorchBackend
    skip_test_backend = False
except ImportError:
    skip_test_backend = True


## Tests


@pytest.mark.skipif(skip_test_backend, reason="PyTorch is not installed.")
def test_set_backend():
    fdtd.set_backend("torch")
    assert isinstance(fdtd.backend, TorchBackend)
    fdtd.set_backend("numpy")
    assert isinstance(fdtd.backend, NumpyBackend)

